-module(kmmind).
-author(hanak@inf.bme.hu).
-vsn('2010-10-26').

%-compile(export_all).
-export([mmind_be/1,mmind_ki/2,megold/2,stopper/2]).

-import(mmind,[mmind/2]).

%% @type code()   = [integer()].
%% @type blacks() = integer().
%% @type whites() = integer().
%% @type answer() = {blacks(),whites()}.
%% @type hint()   = {code(),answer()}.

%% @spec mmind:mmind(Max::int(),Hints::[hint()]) -> Codes::[code()].
%% @doc  Codes a Hints súgás-listának és a Max maximális kódértéknek
%%       megfelelő összes titkos kód tetszőleges sorrendű listája.

%% kmmind-függvények

%% @spec kmmind:mmind_be(FileIn::string()) -> MMind::{integer(),[hint()]}
%% @doc  MMind a FileIn szövegfájlból beolvasott Mmind-feladvány.
%%
mmind_be(FileIn) ->
    case file:open(FileIn,read) of
	{ok, InputFile} ->
	    Max = case io:get_line(InputFile,'') of
		      eof ->
			  io:format("Ures a(z) ~s fajl.~n",[FileIn]),
			  exit('Hibas feladvany');
		      I ->
			  string2integer(cleaned(I))
		  end,
	    Hints = leiro_be(InputFile),
	    file:close(InputFile),
	    case helyes_feladvany(Max,Hints) of
		true ->
		    {Max,Hints};
		false ->
		    io:format("Hibas a feladvany a(z) ~s fajlban.~n",[FileIn]),
		    exit('Hibas feladvany')
	    end;
	_ ->
	    io:format("Nincs ~s nevu fajl.~n",[FileIn]),
	    exit('Nincs ilyen feladvany')
    end.

leiro_be(InputFile) ->
    leiro_be(InputFile,[]).

leiro_be(InputFile,Zss) ->
    case io:get_line(InputFile,'') of
	eof ->
	    lists:reverse(Zss);
	Sor ->
	    Xs = string:tokens(cleaned(Sor)," "),
	    Cs = lists:map(fun string2integer/1,lists:sublist(Xs,length(Xs)-1)),
	    Ys = string:tokens(lists:last(Xs),"/"),
	    [B,W] = lists:map(fun string2integer/1, Ys),
	    case Cs of
		[] ->
		    leiro_be(InputFile,Zss);
		_ ->
		    leiro_be(InputFile,[{Cs,{B,W}}|Zss])
	    end
    end.

cleaned(Sor) ->
    S0 = lists:filter(fun(C) -> C /= $\t end,Sor),
    string:strip(string:strip(S0),right,$\n).

string2integer(S) ->
    element(1,string:to_integer(S)).

%% @spec helyes_feladvany(Max::int(),Hints::[hint()]) -> -> B::bool().
%% @doc  B igaz, ha a feladvány megfelel a specifikációnak.
%%
helyes_feladvany(Max,Hints) ->
    Szinhiba = "A tippekben hibas szinkod van (1 es ~w kozott lehet).~n",
    Tipphiba = "Nem egyforma hosszuak a tippek.~n",
    Valaszhiba = "A valaszok szama hibas (0 es ~w kozott lehet).~n",
    {Ts,As} = lists:unzip(Hints),
    H = length(hd(Ts)),
    Tippteszt = fun(X) -> length(X) == H end,
    Id = fun(X) -> X end,
    Valaszteszt = fun({X,Y}) -> XY = X+Y, 0 =< XY andalso XY =< H end,
    Cs = lists:concat(Ts),
    case
	Max < 1 of
	true ->
	    io:format("A szinek szama ~w; 1-nel nem lehet kisebb.~n",[Max]),
	    false;
	false ->
	    case
		lists:all(Id, lists:map(Tippteszt,Ts)) of
		false ->
		    io:format(Tipphiba),
		    false;
		true ->
		    case
			lists:foldl(fun erlang:min/2, 1, Cs) < 1 orelse
			lists:foldl(fun erlang:max/2, Max, Cs) > Max of
			true ->
			    io:format(Szinhiba,[Max]),
			    false;
			false ->
			    case
				lists:all(Id, lists:map(Valaszteszt,As)) of
				false ->
				    io:format(Valaszhiba,[H]),
				    false;
				true ->
				    true
			    end
		    end
	    end
    end.

%% @spec kmmind:mmind_ki(FileOut::string(),Codes::[code()]) -> void()
%% @doc  A Codes listában átadott Mmind-megoldásokat kiírja
%%       a FileOut szövegfájlba.
%%
mmind_ki(FileOut,SSols) ->
    {ok,OutputFile} = file:open(FileOut,write),
    lists:foreach(fun(Xs) -> io:fwrite(OutputFile,"~w ",[Xs]) end,SSols),
    io:nl(OutputFile),
    file:close(OutputFile).

%% @spec kmmind:megold(FileIn::string(),FileOut::string()) -> void()
%% @doc  Beolvas egy feladványt a FileIn szövegfájlból és összes
%%       megoldását kiírja a FileOut szövegfájlba. Ehhez
%%       felhasználja a mmind:mmind/2 függvényt.
%%
megold(FileIn,FileOut) ->
    {Max,Hints} = mmind_be(FileIn),
    Codes = mmind:mmind(Max,Hints),
    mmind_ki(FileOut,Codes).

%% @spec kmmind:stopper(FileIn::string(),FileOut::string()) -> void()
%% @doc  Mint kmmind:megold/2, de a végén kiírja a FileIn nevét,
%%       a megoldások számát és a futási idõt is.
%%
stopper(FileIn,FileOut) ->
    {Max,Hints} = mmind_be(FileIn),
%    {T, Codes} = timer:tc(mmind,mmind,[MMind]),
    _T = statistics(runtime),
    Codes = mmind:mmind(Max,Hints),
    {_,T} = statistics(runtime),
    mmind_ki(FileOut,Codes),
    io:format("Feladvany: ~s, megoldas: ~3w, futasi ido: ~6w ms\n",
	      [FileIn,length(Codes),T]).
